# create_vector_space.py
# Ambika Acharya and Emily Mallory
# 
# Parameters:
#    -d /path/to/output/directory
#    -t fingerprint type, either maccs (MACCS) or 
#       full (unfolded fingerprint)
#
######################################


reactions_files = open("primary_reactions_data.tsv") # format is rxn_name\tcmpd_A_name\tcmpd_b_name\tsmiles_a\tsmiles_b
drugs_file = open("drug_metabolite_data.tsv") #rxn_name\tdrug_A_name\tmetab_b_name\tsmiles_a\tsmiles_b
compounds_file = open('compounds_to_add.tsv') #format is cmpd_name\tsmiles

from rdkit import DataStructs
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.Fingerprints import FingerprintMols
from rdkit.Chem import UnfoldedRDKFingerprintCountBased
from rdkit.Chem import MACCSkeys
import numpy as np
from numpy import linalg as LA
import sys
import matplotlib
import matplotlib.pyplot as plt
import random
import cPickle as pkl
import argparse

uniq_smiles = set()
smiles_to_reacts = {}
smiles_to_mols = {}
fps = []
fps2 = []
smiles_indices = {}
full_compounds = pkl.load(open("full_compounds_to_smiles.p","rb")) #set of compounds (not just ones in reactions)


'''parse through reaction files, store molecules from smiles codes and reactions'''
def read_input(dump_file):
	'''
	read reaction and compound files for smiles and reactions

	input arguments:
		dump_file: output dictionary of smiles to reactions
	'''

	global uniq_smiles
	print "Starting to read input file..."
	for line in reactions_files:
		data = line.strip().split("\t")
		if data[0] == "reaction":
			continue
		left_mol = data[1]
		right_mol = data[2]
		react = data[0]
		smiles = data[-2:]
		smile_pair = []
		if True:
			for s in smiles:
				if s in skip_comps: continue
				smile_pair.append(s)
				if " " in s or s == "[Mn+++]" or s == "[Fe+++]":
					continue
				else:
					mol = Chem.MolFromSmiles(s)
					if mol is None: #for nitrogens
						continue
					else:
						uniq_smiles.add(s)
						smiles_to_mols[s] = mol
			smiles_to_reacts[tuple(smile_pair)] = react
	print "Finished reading input file"

	count =1
	add =0
	temp_smiles = set()
	for com in full_compounds:
		sm = full_compounds[com]
		if sm in skip_comps: continue
		count+=1

		if " " in sm or sm == "[Mn+++]" or sm == "[Fe+++]":
			continue
		else:
			mol = Chem.MolFromSmiles(sm)
			if mol is None: #for nitrogens
				continue
			else:
				add+=1
				temp_smiles.add(sm)
				smiles_to_mols[sm] = mol
	uniq_smiles = uniq_smiles.union(temp_smiles)

	#add drugs
	for line in drugs_file:
		data = line.strip().split("\t")
		if data[0] == "reaction":
			continue
		react = data[0]
		smile_pair = [data[3],data[4]]
		for s in smile_pair:
			if " " in s or s == "[Mn+++]" or s == "[Fe+++]":
				continue
			else:
				mol = Chem.MolFromSmiles(s)
				if mol is None: #for nitrogens
					continue
				else:
					print react
					uniq_smiles.add(s)
					smiles_to_mols[s] = mol
		smiles_to_reacts[tuple(smile_pair)] = react
	print "Finished reading drugs file"

	#add new compounds
	for line in compounds_file:
		data = line.strip().split("\t")
		print data
		if data[0] == "compound":
			continue
		comp = data[0]
		smiles = data[1]

		if " " in smiles or smiles == "[Mn+++]" or smiles == "[Fe+++]":
			continue
		else:
			mol = Chem.MolFromSmiles(smiles)
			if mol is None: #for nitrogens
				print "skipping " + comp
				continue
			else:
				print smiles
				uniq_smiles.add(smiles)
				smiles_to_mols[smiles] = mol
	print "Finished reading compounds file"
	pkl.dump(smiles_to_reacts, open(dump_file,"wb"))


def get_fps(dump_file, fp_type="maccs"):
	'''
	store fingerprints and smiles codes indexed in the same way

	input arguments:
		fp_type: type of fingerprint, default is maccs
		dump_file: output file for the smiles indices
	'''

	print "Starting to store fingerprints and smiles codes..."
	count = 0

	#### ADDING TO SKIP LOAD FUNCTION ####
	uniq_smiles2 = pkl.load(open("uniq_smiles_final_results_maccs.p"))

	for s in uniq_smiles2:
		mol = Chem.MolFromSmiles(s)
		if mol is None: #for nitrogens
			continue
		else:
			smiles_to_mols[s] = mol

	#### END ####

	for s in uniq_smiles2:
		print s
		fp = MACCSkeys.GenMACCSKeys(smiles_to_mols[s])
		fps2.append(fp)
		if fp_type == "full":
			fp = UnfoldedRDKFingerprintCountBased(smiles_to_mols[s])#use function to get full unfolded
			if not fp.GetNonzeroElements():
				continue
		fps.append(fp) #use function to get full unfolded
		smiles_indices[s] = count
		count +=1
	pkl.dump(smiles_indices, open(dump_file,"wb"))
	print "Finished storing fingerprints and smiles codes"

def calculate_similarites(dump_file, fp_type="maccs"):
	'''
	Calculate similarity matrix for all compounds

	input arguments:
		fp_type: type of fingerprint, default is maccs
		dump_file: output file for the similarity matrix
	'''

	print "Starting to generate similarities table..."
	similarities = [[0 for x in fps] for x in fps]
	plot_sims = []
	plot_sims2 = []
	for i in range(len(fps)):
		for j in range(len(fps)):
			sim = 0
			if fp_type == "full":
				union = len(set(fps[i].GetNonzeroElements().keys()).union(set(fps[j].GetNonzeroElements().keys())))
				intersection = len(set(fps[i].GetNonzeroElements().keys()).intersection(set(fps[j].GetNonzeroElements().keys())))
				sim = intersection/float(union)
			else:
				sim = DataStructs.FingerprintSimilarity(fps[i],fps[j])
			similarities[i][j] = sim
	print "Finished generating similarities table"
	pkl.dump(similarities, open(dump_file,"wb"))
	return similarities

def do_eigen_decomp(similarities,dump_file):
	'''
	perform eigen decomposition

	input arguments:
		similairities: similarity matrix
		dump_file: output file for the vector space
	'''

	print "Starting eigen decomp..."

	D, V = LA.eigh(np.array(similarities))
	V_fin = V[:,-50:] #or V[:,-8:] for top vals #specify dimensionality reduction
	D_fin = D[-50:] #or D[:,-8:] for top vals #specify dimensionality reduction
	for i in range(len(D_fin)):
		if D_fin[i] < .000001:
			D_fin[i] = 0.0
	#look at the distribution 10^-4
	D_half = np.sqrt(D_fin)
	D_half = np.diag(D_half)
	X_mol = np.dot(V_fin,D_half)
	print "Finished eigen decomp"
	pkl.dump(X_mol, open(dump_file,"wb"))


def get_arguments():
	'''
	Name: get_arguments
	Parses arguments from command line

	argparse arguments:
		-d: output directory
		-t: fingerprint type, either maccs or full (unfolded fingerprint)
	'''

	parser = argparse.ArgumentParser()

	parser.add_argument('-d', action='store', dest='dir',
                    help='output directory')

	parser.add_argument('-t', action='store', dest='fp_type',
                    help='store type of fingerprint: full or maccs')

	return parser.parse_args()


def main():

	# get command line arguments
	# -d output directory name
	# -t fingerprint type: either maccs or full (unfolded fp)
	arguments = get_arguments()

	dir_name = arguments.dir
	fp_type = arguments.fp_type

	print dir_name, fp_type
	
	# calculate fingerprints for each molecule
	get_fps(dir_name + "/smiles_indices_"+fp_type+".p", fp_type)

	# calculate the similarity matrix using Tanimoto coefficient
	sims = calculate_similarites(dir_name + "/similarities_"+fp_type+".p", fp_type)

	# perform eigen decomposition using similarity matrix. Output vector space
	do_eigen_decomp(sims,dir_name + "/X_mol_50_"+fp_type+".p")

if __name__ == '__main__':
	main()
